# Copyright 2014 Sardar Yumatov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__title__ = 'APNS client'
__version__ = "0.2"
__author__ = "Sardar Yumatov"
__contact__ = "ja.doma@gmail.com"
__license__ = "Apache 2.0"
__homepage__ = "https://bitbucket.org/sardarnl/apns-client/"
__copyright__ = 'Copyright 2014 Sardar Yumatov'


from apnsclient.apns import APNs, Message
from apnsclient.transport import Session
